#
# fill_rdoc_comment.rb
#
# Copyright 2005-2006 Tero Hasu
#

$debug = ($0 == __FILE__)

opts = ARGV.shift.split(/,/)
cmtstyle = opts.first

# Whether to keep any indentation appearing after any comment
# character on every line of input.
$keep_indent = opts.include? "indent"

$cmts =
  case cmtstyle
  when "none"
    nil
  when "ruby"
    '#'
  when "tex"
    "%"
  when "lisp"
    ";"
  when "cxx"
    "//"
  when "haskell"
    "--"
  else
    raise
  end

if $cmts
  # Collects the spaces before any comment string.
  $level_re = Regexp.new("\\A(\\s*)" + Regexp.escape($cmts))

  # Collects the comment characters. The last character may repeat.
  $prefix_re = Regexp.new("^\\s*(" + Regexp.escape($cmts) + "+)")

  # Matches comment characters, and any space before them.
  $prefix_res = "^(\\s*(?:" + Regexp.escape($cmts) + "+))"
else
  $prefix_re = $level_re = /^()/
  $prefix_res = "^()"
end

require 'pp'
require 'strscan'

require 'emacs/rdoc_fill'

require 'codeprint'
include CodePrint

if $debug
  data = File.read(File.join(File.dirname(File.expand_path(__FILE__)), "rdoctestdata"))
else
  data = $stdin.read
end

def log(data)
  File.open('/tmp/output', 'w') do |output|
    PP::pp(data, output)
  end
end

# Like String.size, but accounts for Tabs, assuming Tab width 8.
def text_width(text, initlv = 0)
  level = initlv
  input = StringScanner.new text
  until input.eos?
    if input.scan(/\t/)
      level += 1
      level += 1 until (level % 8) == 0
    else
      input.getch
      level += 1
    end
  end
  return level
end

# If we are to retain any indentation appearing after the comment and
# before the RDoc text, we must know how much to retain. We calculate
# the width of the indentation for all non-empty lines in the region,
# and take the smallest width.
if $keep_indent
  $keepsp = nil
  emptyre = /^\s*$/
  re = Regexp.new($prefix_res + "(\\s*)")
  data.each_line do |line|
    if line !~ emptyre and line =~ re
      cmspace = $1
      inspace = $2
      thissp = text_width(inspace, text_width(cmspace))
      $keepsp = thissp if (!$keepsp) || (thissp < $keepsp)
    end
  end
  $keepsp ||= 0
end

if (data =~ $level_re)
  if $cmts
    level = text_width($1)
    data =~ $prefix_re or raise data.inspect
    inds = [space(level), $1]

    # Remove comment prefixes to keep them from confusing the RDoc
    # parser.
    data.gsub!($prefix_re, "")
  else
    inds = []
  end

  if $keep_indent
    inds.push(space($keepsp))
  elsif $cmts
    inds.push(space)
  end

  rdoc = RDocFormatter::new(data)
  rdoc.parse
  #pp rdoc.lines
  #log(rdoc.lines)

  data = ""
  # Emacs' fill-column defaults to 70; i.e. filling happens
  # _beyond_ column 70. Not nice to hardcode the default here,
  # so please someone fix this (can be passed as an argument
  # to this script).
  q = CodePp.new(:output => data, :maxwidth => 70)
  q.indent(inds) do
    rdoc.unparse(q)
  end
end

#log(data)
$stdout.write(data)
